package com.comtika.safe.common.request;



import com.comtika.safe.common.utils.StringUtil;

import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;


/**
 * Request处理工具
 * @since JDK1.8
 * @version V1.0.0 2019-02-15
 * @author zly [494977551@qq.com]
 *
 */
public class RequestUtil {

	/**
	 * 获取String类型参数,若无该输入参数，则返回NULL
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @return String型的输入参数
	 */
	public static String getStringParam(ServletRequest request, String paramName) {
		return getStringParam(request, paramName, null);
	}

	/**
	 * 获取String类型参数
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @param defaults
	 *            设定的默认值
	 * @return String型的输入参数
	 */
	public static String getStringParam(ServletRequest request, String paramName, String defaults) {
		String str = request.getParameter(paramName);
		return StringUtil.convertString(str, defaults);
	}

	/**
	 * 获取int类型参数，若无该输入参数，则返回NULL
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @return Integer型的输入参数
	 */
	public static Integer getIntParam(ServletRequest request, String paramName) {
		return getIntParam(request, paramName, null);
	}

	/**
	 * 获取int类型参数
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @param defaults
	 *            设定的默认值
	 * @return Integer型的输入参数
	 */
	public static Integer getIntParam(ServletRequest request, String paramName, Integer defaults) {
		if(request != null){
			return StringUtil.convertInt(request.getParameter(paramName), defaults);
		}
		return defaults;
	}

	/**
	 * 获取long类型参数，若无该输入参数，则返回NULL
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @return Long型的输入参数
	 */
	public static Long getLongParam(ServletRequest request, String paramName) {
		return getLongParam(request, paramName, null);
	}

	/**
	 * 获取long类型参数
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @param defaults
	 *            设定的默认值
	 * @return Long型的输入参数
	 */
	public static Long getLongParam(ServletRequest request, String paramName, Long defaults) {
		if(request != null){
			return StringUtil.convertLong(request.getParameter(paramName), defaults);
		}
		return defaults;
	}

	/**
	 * 获取double类型参数，若无该参数，则返回NULL
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @return Double型的输入参数
	 */
	public static Double getDoubleParam(ServletRequest request, String paramName) {
		return getDoubleParam(request, paramName, null);
	}

	/**
	 * 获取double类型参数
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @param defaults
	 *            设定的默认值
	 * @return Double型的输入参数
	 */
	public static Double getDoubleParam(ServletRequest request, String paramName, Double defaults) {
		if(request != null){
			return StringUtil.convertDouble(request.getParameter(paramName), defaults);
		}
		return defaults;
	}

	/**
	 * 获取short类型参数，若无该参数，则返回NULL
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @return Short型的输入参数
	 */
	public static Short getShortParam(ServletRequest request, String paramName) {
		return getShortParam(request, paramName, null);
	}

	/**
	 * 获取short类型参数
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @param defaults
	 *            设定的默认值
	 * @return Short型的输入参数
	 */
	public static Short getShortParam(ServletRequest request, String paramName, Short defaults) {
		return StringUtil.convertShort(request.getParameter(paramName), defaults);
	}

	/**
	 * 获取float类型参数，若无该参数，则返回NULL
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @return Float型的输入参数
	 */
	public static Float getFloatParam(ServletRequest request, String paramName) {
		return getFloatParam(request, paramName, null);
	}

	/**
	 * 获取float类型参数
	 *
	 * @param request
	 *            ServletRequest的实例
	 * @param paramName
	 *            参数名字
	 * @param defaults
	 *            设定的默认值
	 * @return Float型的输入参数
	 */
	public static Float getFloatParam(ServletRequest request, String paramName, Float defaults) {
		if(request != null){
			return StringUtil.convertFloat(request.getParameter(paramName), defaults);
		}
		return defaults;
	}

	/**
	 * 获取boolean类型的参数，若无该参数，默认值为NULL
	 *
	 * @param request
	 * @param paramName
	 *
	 * @return Boolean
	 */
	public static Boolean getBooleanParam(ServletRequest request, String paramName) {
		return getBooleanParam(request, paramName, null);
	}

	/**
	 * 获取boolean类型的参数
	 *
	 * @param request
	 * @param paramName
	 * @param defaults
	 *
	 * @return Boolean
	 */
	public static Boolean getBooleanParam(ServletRequest request, String paramName, Boolean defaults) {
		if(request != null){
			return StringUtil.convertBoolean(request.getParameter(paramName), defaults);
		}
		return defaults;

	}

	/**
	 * 获取所有参数的MAP集合
	 *
	 * @param request
	 * @param clazz
	 *
	 * @return MAP
	 */
	public static Map<String, String> getParamMap(ServletRequest request){
		ConcurrentHashMap<String, String> paramMap = new ConcurrentHashMap<>();
		Enumeration<String> names = request.getParameterNames();
		while (names.hasMoreElements()) {
			String key = names.nextElement();
			String value = request.getParameter(key);
			paramMap.put(key, value);
		}
		return paramMap;
	}

	/**
	 * 获取所有参数转换为JAVABEAN
	 *
	 * @param request
	 * @param clazz
	 *
	 * @return T
	 */
	public static <T> T getParamBean(ServletRequest request, Class<T> clazz){
		Map<String, String> paramMap = getParamMap(request);
		try {
			T obj = clazz.newInstance();
			BeanInfo info = Introspector.getBeanInfo(clazz);
			PropertyDescriptor[] pds = info.getPropertyDescriptors();
			for (PropertyDescriptor property : pds) {
				String name = property.getName();
				if(paramMap.containsKey(name)){
					try {
						Method setter = property.getWriteMethod();
						String val = paramMap.get(name);
						if(StringUtil.isEmpty(val)){
							continue;
						}
						Class<?> type = property.getPropertyType();
						if(int.class.isAssignableFrom(type) || Integer.class.isAssignableFrom(type)){
							setter.invoke(obj, Integer.parseInt(val));
						}
						else if(long.class.isAssignableFrom(type) || Long.class.isAssignableFrom(type)){
							setter.invoke(obj, Long.parseLong(val));
						}
						else if(double.class.isAssignableFrom(type) || Double.class.isAssignableFrom(type)){
							setter.invoke(obj, Double.parseDouble(val));
						}
						else if(float.class.isAssignableFrom(type) || Float.class.isAssignableFrom(type)){
							setter.invoke(obj, Float.parseFloat(val));
						}
						else if(short.class.isAssignableFrom(type) || Short.class.isAssignableFrom(type)){
							setter.invoke(obj, Short.parseShort(val));
						}
						else if(boolean.class.isAssignableFrom(type) || Boolean.class.isAssignableFrom(type)){
							setter.invoke(obj, Boolean.parseBoolean(val));
						}
						else if(byte.class.isAssignableFrom(type) || Byte.class.isAssignableFrom(type)){
							setter.invoke(obj, Byte.parseByte(val));
						}
						else{
							setter.invoke(obj, val);
						}
					} catch (Exception e) {
						System.out.println("The Request Param Convert To Field Fail: ["+name+"]");
						e.printStackTrace();
					}
				}
			}
			return obj;
		} catch (InstantiationException | IllegalAccessException | IntrospectionException e) {
			throw new RuntimeException("The Class Instance Fail By Reflect: ["+clazz.getName()+"]", e);
		}
	}



	/**
	 * 获取request的属性
	 *
	 * @param request
	 * @param attrName
	 *
	 * @return Object
	 */
	public static Object getAttribute(ServletRequest request, String attrName) {
		return request.getAttribute(attrName);
	}

	/**
	 * 获取request中的属性,转换为JAVA对象
	 *
	 * @param request
	 * @param attrName
	 * @param clazz
	 *
	 * @return NULL or T
	 */
	@SuppressWarnings("unchecked")
	public static <T> T getAttribute(ServletRequest request, String attrName, Class<T> clazz) {
		Object obj = request.getAttribute(attrName);
		if(null == obj){
			return null;
		}
		Class<?> cls = obj.getClass();
		if(clazz.isAssignableFrom(cls)){
			return (T) obj;
		}else{
			throw new ClassCastException(cls.getName()+" cannot be cast to "+clazz.getName());
		}
	}

	/**
	 * 获取String类型的属性
	 *
	 * @param request
	 * @param attrName
	 *
	 * @return
	 */
	public static String getStringAttr(ServletRequest request, String attrName) {
		return (String) getAttribute(request, attrName);
	}

	/**
	 * 获取int类型的属性
	 *
	 * @param request
	 * @param attrName
	 *
	 * @return
	 */
	public static Integer getIntAttr(ServletRequest request, String attrName) {
		return (Integer) getAttribute(request, attrName);
	}

	/**
	 * 获取long类型的属性
	 *
	 * @param request
	 * @param attrName
	 *
	 * @return
	 */
	public static Long getLongAttr(ServletRequest request, String attrName) {
		return (Long) getAttribute(request, attrName);
	}

	/**
	 * 获取double类型的属性
	 *
	 * @param request
	 * @param attrName
	 *
	 * @return
	 */
	public static Double getDoubleAttr(ServletRequest request, String attrName) {
		return (Double) getAttribute(request, attrName);
	}

	/**
	 * 获取float类型的属性
	 *
	 * @param request
	 * @param attrName
	 *
	 * @return
	 */
	public static Float getFloatAttr(ServletRequest request, String attrName) {
		return (Float) getAttribute(request, attrName);
	}

	/**
	 * 获取boolean类型的属性
	 *
	 * @param request
	 * @param attrName
	 *
	 * @return
	 */
	public static Boolean getAttributeBoolean(ServletRequest request, String attrName) {
		return (Boolean) getAttribute(request, attrName);
	}


	/**
	 * 获取RequestURI
	 *
	 * @param req
	 * @param skipContextPath 是否过滤ContextPath路径
	 *
	 * @return
	 */
	public static String getRequestURI(HttpServletRequest req, boolean skipContextPath){
        String requestUri = req.getRequestURI().trim();
        if(skipContextPath){
        	String contextPath = req.getContextPath().trim();
        	return requestUri.replaceFirst(contextPath, "");
        }
        return requestUri;
	}


	/**
	 * 获取用户地址
	 *
	 * @param request
	 * @return
	 *
	 */
	public static String getRemoteAddr(HttpServletRequest request){
		String ip = request.getHeader("X-Real-IP");
		if(ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)){
			ip = request.getHeader("Proxy-Client-IP");
		}
		if(ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)){
			ip = request.getHeader("WL-Proxy-Client-IP");
		}
		if(ip == null || ip.length() == 0 || "unknown".equalsIgnoreCase(ip)){
			ip = request.getRemoteAddr();
		}

		ip = ip.equals("0:0:0:0:0:0:0:1") ? "127.0.0.1" : ip;

		int index = ip.indexOf(",");
		if (index != -1) {
			ip = ip.substring(0, index);
		}
		return ip;
	}


}
