package com.comtika.safe.common.utils;


import java.io.*;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * @desc	:	字符串工具
 * @author	:	actorLucky[494977551@qq.com]
 * @date	:	2018年3月31日 上午11:20:28
 */
public class StringUtil {


	/** UTF-8编码常量 */
	public static final String ENC_UTF8 = "UTF-8";
	/** GBK编码常量 */
	public static final String ENC_GBK = "GBK";
	/** GBK的Charset */
	public static final Charset GBK = Charset.forName("GBK");
	/** UTF-8的Charset */
	public static final Charset UTF_8 = Charset.forName("UTF-8");


	/**
	 * 过滤SQL字符串,防止SQL inject
	 *
	 * @param sql
	 *
	 * @return String
	 */
	public static String encodeSQL(String sql) {
		if (sql == null) {
			return "";
		}
		// 不用正则表达式替换，直接通过循环，节省cpu时间
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < sql.length(); ++i) {
			char c = sql.charAt(i);
			switch (c) {
			case '\\':
				sb.append("\\\\");
				break;
			case '\r':
				sb.append("\\r");
				break;
			case '\n':
				sb.append("\\n");
				break;
			case '\t':
				sb.append("\\t");
				break;
			case '\b':
				sb.append("\\b");
				break;
			case '\'':
				sb.append("\\\'");
				break;
			case '\"':
				sb.append("\\\"");
				break;
			case '\u200B':// ZERO WIDTH SPACE
			case '\uFEFF':// ZERO WIDTH NO-BREAK SPACE
				break;
			default:
				sb.append(c);
			}
		}
		return sb.toString();
	}


	/**
	 * 转化字符型参数，若输入字符串为空，则返回设定的默认值
	 *
	 * @param str
	 *            输入字符串
	 * @param defaults
	 *            默认值
	 * @return 字符串参数
	 */
	public static String convertString(String str, String defaults) {
		if (isEmpty(str)) {
			return defaults;
		} else {
			return str;
		}
	}

	/**
	 * 转换int参数，若输入字符串为null或不能转为int，则返回设定的默认值
	 *
	 * @param str
	 *            输入字符串
	 * @param defaults
	 *            默认值
	 * @return int参数
	 */
	public static Integer convertInt(String str, Integer defaults) {
		if (isEmpty(str)) return defaults;
		try {
			return Integer.parseInt(str.trim());
		} catch (Exception e) {
			return defaults;
		}
	}

	/**
	 * 转换long型参数，若输入字符串为null或不能转为long，则返回设定的默认值
	 *
	 * @param str
	 *            输入字符串
	 * @param defaults
	 *            默认值
	 * @return long参数
	 */
	public static Long convertLong(String str, Long defaults) {
		if (isEmpty(str)) {
			return defaults;
		}
		try {
			return Long.parseLong(str.trim());
		} catch (Exception e) {
			return defaults;
		}
	}

	/**
	 * 转换double型参数，若输入字符串为null或不能转为double，则返回设定的默认值
	 *
	 * @param str
	 *            输入字符串
	 * @param defaults
	 *            默认值
	 * @return double型参数
	 */
	public static double convertDouble(String str, double defaults) {
		if (isEmpty(str)) {
			return defaults;
		}
		try {
			return Double.parseDouble(str.trim());
		} catch (Exception e) {
			return defaults;
		}
	}

	/**
	 * 转换short型参数，若输入字符串为null或不能转为short，则返回设定的默认值
	 *
	 * @param str
	 *            输入字符串
	 * @param defaults
	 *            默认值
	 * @return short型参数
	 */
	public static short convertShort(String str, short defaults) {
		if (isEmpty(str)) {
			return defaults;
		}
		try {
			return Short.parseShort(str.trim());
		} catch (Exception e) {
			return defaults;
		}
	}

	/**
	 * 转换float型参数，若输入字符串为null或不能转为float，则返回设定的默认值
	 *
	 * @param str
	 *            输入字符串
	 * @param defaults
	 *            默认值
	 * @return float型参数
	 */
	public static float convertFloat(String str, float defaults) {
		if (isEmpty(str)) {
			return defaults;
		}
		try {
			return Float.parseFloat(str.trim());
		} catch (Exception e) {
			return defaults;
		}
	}

	/**
	 * 转换boolean型参数，若输入字符串为null或不能转为boolean，则返回设定的默认值
	 *
	 * @param str
	 *            输入字符串
	 * @param defaults
	 *            默认值
	 * @return boolean型参数
	 */
	public static boolean convertBoolean(String str, boolean defaults) {
		if (isEmpty(str)) {
			return defaults;
		}
		try {
			return Boolean.parseBoolean(str.trim());
		} catch (Exception e) {
			return defaults;
		}
	}

	/**
	 * 分割字符串
	 *
	 * @param line
	 *            原始字符串
	 * @param seperator
	 *            分隔符
	 * @return 分割结果
	 */
	public static String[] split(String line, String seperator) {
		if (line == null || seperator == null || seperator.length() == 0) {
			return new String[0];
		}
		ArrayList<String> list = new ArrayList<String>();
		int pos1 = 0;
		int pos2;
		for (;;) {
			pos2 = line.indexOf(seperator, pos1);
			if (pos2 < 0) {
				list.add(line.substring(pos1));
				break;
			}
			list.add(line.substring(pos1, pos2));
			pos1 = pos2 + seperator.length();
		}
		// 去掉末尾的空串，和String.split行为保持一致
		for (int i = list.size() - 1; i >= 0; --i) {
			if(list.get(i).length() == 0){
				list.remove(i);
			}
		}
		return list.toArray(new String[0]);
	}

	/**
	 * 分割字符串，并转换为int数组
	 *
	 * @param line
	 *            原始字符串
	 * @param seperator
	 *            分隔符
	 * @param def
	 *            默认值,无法转换的元素会用该值代替
	 * @return 分割结果
	 */
	public static int[] splitInt(String line, String seperator, int def) {
		String[] ss = split(line, seperator);
		int[] r = new int[ss.length];
		for (int i = 0; i < r.length; ++i) {
			r[i] = convertInt(ss[i], def);
		}
		return r;
	}

	/**
	 * 分割字符串，并转换为long数组
	 *
	 * @param line
	 *            原始字符串
	 * @param seperator
	 *            分隔符
	 * @param def
	 *            默认值,无法转换的元素会用该值代替
	 * @return 分割结果
	 */
	public static long[] splitLong(String line, String seperator, long def) {
		String[] ss = split(line, seperator);
		long[] r = new long[ss.length];
		for (int i = 0; i < r.length; ++i) {
			r[i] = convertLong(ss[i], def);
		}
		return r;
	}

	/**
	 * 分割字符创,大写每段首字母后在拼接
	 *
	 * @param line 原字符串
	 *
	 * @param seperator 分隔符
	 * @return
	 */
	public static String splitToUpperCase(String line, String seperator) {
		String[] array = split(line, seperator);
		if (array == null || array.length < 1) {
			return line;
		}
//		String str = "";
		StringBuffer stringBuffer = new StringBuffer();
		for (String s : array) {
			stringBuffer.append(s);
//			str += toUpperCaseFirst(s);
		}
		return stringBuffer.toString();
	}

	/**
	 * 字符串全量替换
	 *
	 * @param line
	 *            原始字符串
	 * @param src
	 *            需要被替换的内容
	 * @param dest
	 *            替换后的内容
	 * @return 结果
	 */
	public static String replaceAll(String line, String src, String dest) {
		if (line == null || src == null || dest == null || src.length() == 0){
			return line;
		}
		int pos = line.indexOf(src); // 查找第一个替换的位置
		if (pos < 0){
			return line;
		}
		int capacity = dest.length() > src.length() ? line.length() * 2 : line.length();
		StringBuilder sb = new StringBuilder(capacity);
		int writen = 0;
		for (; pos >= 0;) {
			sb.append(line, writen, pos); 		// append 原字符串不需替换部分
			sb.append(dest); 					// append 新字符串
			writen = pos + src.length(); 		// 忽略原字符串需要替换部分
			pos = line.indexOf(src, writen); 	// 查找下一个替换位置
		}
		sb.append(line, writen, line.length()); // 替换剩下的原字符串
		return sb.toString();
	}

	/**
	 * 只替换第一个字符串
	 *
	 * @param line
	 *            原始字符串
	 * @param src
	 *            需要被替换的内容
	 * @param dest
	 *            替换后的内容
	 * @return 结果
	 */
	public static String replaceFirst(String line, String src, String dest) {
		if (line == null || src == null || dest == null || src.length() == 0){
			return line;
		}
		int pos = line.indexOf(src);
		if (pos < 0) {
			return line;
		}
		StringBuilder sb = new StringBuilder(line.length() - src.length() + dest.length());

		sb.append(line, 0, pos);
		sb.append(dest);
		sb.append(line, pos + src.length(), line.length());
		return sb.toString();
	}

	/**
	 * 只替换最后一个字符串
	 *
	 * @param line
	 *            原始字符串
	 * @param src
	 *            需要被替换的内容
	 * @param dest
	 *            替换后的内容
	 * @return 结果
	 */
	public static String replaceLast(String line, String src, String dest) {
		if (line == null || src == null || dest == null || src.length() == 0){
			return line;
		}
		int pos = line.lastIndexOf(src);
		if (pos < 0) {
			return line;
		}
		StringBuilder sb = new StringBuilder(line.length() - src.length() + dest.length());

		sb.append(line, 0, pos);
		sb.append(dest);
		sb.append(line, pos + src.length(), line.length());
		return sb.toString();
	}

	/**
	 * 删除指定的全部字符串
	 *
	 * @param line
	 *            原字符串
	 * @param del
	 *            需要删除的内容
	 *
	 * @return
	 */
	public static String removeAll(String line, String del) {
		return replaceAll(line, del, "");
	}

	/**
	 * 删除指定的第一个字符串
	 *
	 * @param line
	 *            原字符串
	 * @param del
	 *            需要删除的内容
	 *
	 * @return
	 */
	public static String removeFirst(String line, String del) {
		return replaceFirst(line, del, "");
	}

	/**
	 * 删除指定的最后一个字符串
	 *
	 * @param line
	 *            原字符串
	 * @param del
	 *            需要删除的内容
	 *
	 * @return
	 */
	public static String removeLast(String line, String del) {
		return replaceLast(line, del, "");
	}

	/**
	 * 大写字符串指定下标的字符
	 *
	 * @param line 原字符串
	 *
	 * @param index 需要大写的下标
	 * @return
	 */
	public static String toUpperCase(String line, int index) {
		char[] cs = line.toCharArray();
		char c = cs[index];
		if (c >= 97 && c <= 122) {
			cs[index] -= 32;
		}
		return String.valueOf(cs);
	}

	/**
	 * 大写字符串首字母
	 * @param line
	 * @return
	 */
	public static String toUpperCaseFirst(String line) {
		return toUpperCase(line, 0);
	}

	/**
	 * 小写字符串指定下标的字符
	 *
	 * @param line 原字符串
	 *
	 * @param index 需要小写的下标
	 * @return
	 */
	public static String toLowerCase(String line, int index) {
		char[] cs = line.toCharArray();
		char c = cs[index];
		if (c >= 35 && c <= 90) {
			cs[index] += 32;
		}
		return String.valueOf(cs);
	}

	/**
	 * 小写字符串首字母
	 *
	 * @param line 原字符串
	 * @return
	 */
	public static String toLowerCaseFirst(String line) {
		return toLowerCase(line, 0);
	}

	/**
	 * 判断字符串为空 [null 或 s.trim().isEmpty()]
	 *
	 * @param str
	 *
	 * @return
	 */
	public static boolean isEmpty(String str) {
		if (str == null || "null".equals(str)||"".equals(str)){
			return true;
		}
		return str.trim().isEmpty();
	}

	/**
	 * 判断字符串不为空
	 *
	 * @param str
	 *
	 * @return
	 */
	public static boolean isNotEmpty(String str) {
		return !isEmpty(str);
	}

	/***
	 * 判断字符串数组不为空
	 * @param strs
	 * @return
	 */
	public static boolean isNotEmpty(String...strs) {
		for(String str:strs)
			if(isEmpty(str)) return false;
		return true;
	}

	/**
	 * 获取字符串的UTF-8编码字节数组
	 *
	 * @param s
	 * @return
	 */
	public static byte[] getUTF8Bytes(String s) {
		if (s != null && s.length() >= 0) {
			return s.getBytes(UTF_8);
		}
		return null;
	}

	/**
	 * 获取字符串的GBK编码字节数组
	 *
	 * @param s
	 * @return
	 */
	public static byte[] getGBKBytes(String s) {
		if (s != null && s.length() >= 0) {
			return s.getBytes(GBK);
		}
		return null;
	}

	/**
	 * 获取字节数组的UTF-8编码字符串
	 *
	 * @param b
	 * @return
	 */
	public static String getUTF8String(byte[] b) {
		if (b != null) {
			return new String(b, UTF_8);
		}
		return null;
	}

	/**
	 * 获取字节数组的GBK编码字符串
	 *
	 * @param b
	 * @return
	 */
	public static String getGBKString(byte[] b) {
		if (b != null) {
			return new String(b, GBK);
		}
		return null;
	}

	/**
	 * 对字符串以 GBK编码方式进行URLEncode
	 *
	 * @param s
	 * @return
	 */
	public static String URLEncodeGBK(String s) {
		if (s != null && s.length() > 0) {
			try {
				return URLEncoder.encode(s, ENC_GBK);
			} catch (UnsupportedEncodingException e) {
			}
		}
		return s;
	}

	/**
	 * 对字符串以 UTF-8编码方式进行URLEncode
	 *
	 * @param s
	 * @return
	 */
	public static String URLEncodeUTF8(String s) {
		if (s != null && s.length() > 0) {
			try {
				return URLEncoder.encode(s, ENC_UTF8);
			} catch (UnsupportedEncodingException e) {
			}
		}
		return s;
	}

	/**
	 * 对字符串以 GBK编码方式进行URLDecode
	 *
	 * @param s
	 * @return
	 */
	public static String URLDecodeGBK(String s) {
		if (s != null && s.length() > 0) {
			try {
				return URLDecoder.decode(s, ENC_GBK);
			} catch (UnsupportedEncodingException e) {
			}
		}
		return s;
	}

	/**
	 * 对字符串以 UTF-8编码方式进行URLDecode
	 *
	 * @param s
	 * @return
	 */
	public static String URLDecodeUTF8(String s) {
		if (s != null && s.length() > 0) {
			try {
				return URLDecoder.decode(s, ENC_UTF8);
			} catch (UnsupportedEncodingException e) {
			}
		}
		return s;
	}
/*	public static String join(Collection collection, String separator) {
		if (collection == null)return null;
		return join(collection.iterator(), separator);
	}*/
	@SuppressWarnings("unused")
	private static final String EMPTY = "";
	/*public static String join(Iterator iterator, String separator) {
		if (iterator == null) return null;
		if (!iterator.hasNext()) return EMPTY;
		Object first = iterator.next();
		if (!iterator.hasNext()) return ObjectUtils.toString(first);
		StrBuilder buf = new StrBuilder(256);
		if (first != null) buf.append(first);
		while (iterator.hasNext()) {
			if (separator != null)buf.append(separator);
			Object obj = iterator.next();
			if (obj != null) buf.append(obj);
		}
		return buf.toString();
	}*/
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public static Map<String, String> toStringMap(String... pairs) {
		Map<String, String> parameters = new HashMap();
		if (pairs.length > 0) {
			if (pairs.length % 2 != 0) {
				throw new IllegalArgumentException("pairs must be even.");
			}

			for(int i = 0; i < pairs.length; i += 2) {
				parameters.put(pairs[i], pairs[i + 1]);
			}
		}

		return parameters;
	}
	public static Map<String,Object> toMap(Object...pairs){
		Map<String,Object> res = new HashMap<>();
		String key;
		if(pairs != null && pairs.length != 0){
			int len = pairs.length;
			if(len % 2 != 0){
				throw new IllegalArgumentException("toMap pairs can not be odd number.");
			}
			for (int i=0;i<len;i+=2){
				key = pairs[i].toString();
				res.put( key, pairs[i+1]);
			}
			return res;
		} else return res;
	}
	public static boolean isEmpty(Collection<?> collection) {
		return collection == null || collection.isEmpty();
	}

	public static boolean isNotEmpty(Collection<?> collection) {
		return collection != null && !collection.isEmpty();
	}

	public static void main(String[] args)throws Exception {
		FileReader in = new FileReader("F:/xzzf.txt");
		String read = read(in);
		System.out.println(read);
	}

	public static String read(Reader reader) throws IOException {
		StringWriter writer = new StringWriter();

		String var2;
		try {
			write((Reader)reader, (Writer)writer);
			var2 = writer.getBuffer().toString();
		} finally {
			writer.close();
		}

		return var2;
	}
	public static long write(Reader reader, Writer writer) throws IOException {
		return write((Reader)reader, (Writer)writer, 8192);
	}
	public static long write(Reader reader, Writer writer, int bufferSize) throws IOException {
		long total = 0L;

		int read;
		for(char[] buf = new char[8192]; (read = reader.read(buf)) != -1; total += (long)read) {
			writer.write(buf, 0, read);
		}

		return total;
	}


}
