package com.comtika.safe.core.protocol.ota;

import com.comtika.safe.common.enums.equip.terminal.TerminalStatusEnum;
import com.comtika.safe.common.response.RespResult;
import com.comtika.safe.common.response.RespWrapper;
import com.comtika.safe.common.utils.StringUtil;
import com.comtika.safe.core.protocol.cache.LocalProtocolCache;
import com.comtika.safe.core.protocol.cache.ProtocolRedisCache;
import com.comtika.safe.core.protocol.entity.netty.ResProtocol;
import com.comtika.safe.core.protocol.entity.version.OTAUpgradeEntity;
import com.comtika.safe.core.protocol.eunms.*;
import com.comtika.safe.core.protocol.entity.netty.OrderEntity;
import com.comtika.safe.core.protocol.eunms.netty.OperationCode;
import com.comtika.safe.core.protocol.eunms.netty.OrderCodeType;
import com.comtika.safe.core.protocol.eunms.version.OTAStatus;
import com.comtika.safe.core.protocol.responseUtil.FutureAdapter;
import com.comtika.safe.core.protocol.utils.SocketUtil;
import com.comtika.safe.dao.equip.terminal.EquipTerminalDao;
import com.comtika.safe.entity.equip.terminal.CtkEquipTerminal;
import io.netty.channel.Channel;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import static com.comtika.safe.core.protocol.ProtocolConstant.RETRY_NUM;


/***
 * @ClassName: ThreadHandleOTA
 * @Description: 多线程处理ota升级问题 指令
 * 		一条指令有3次重试，等待硬件响应时间为30s
 * 		ota指令有：
 * 			Server读取Client可用固件存储空间 A0H
 * 			Server通知Client即将开始固件传输 A1H
 * 			Server传输给Client固件的指令 A2H
 * 			Server传输固件结束 A3H
 *    		Server询问Client固件校验是否正确 A4H
 *          Server要求Client进行固件自更新 A5H
 * @author zhaoLuYang
 * @date2018年8月17日 上午9:17:07
 * @version V1.0
 * 优化：2018/12/05 代码质量优化、去掉端点升级功能
 */
@Component
@Log4j2
public class ThreadHandleOTA {
	private static final byte[] DEFAULT_BYTE = new byte[] {};
	//升级包只在发送大小
	private static final int LENGTH = 1024;
	//只有测试环境debug 有用
	private static final boolean DEBUG = true;

	/** 集中器 **/
	@Autowired
	private EquipTerminalDao terminalDao;
	/**端点缓存*/
	@Autowired
	private ProtocolRedisCache protocolRedisCache;



	@Async
	public void run(ProtocolOTAEntity pt) {
		CtkEquipTerminal terminal = pt.getTerminal();
		if(terminal == null)return;
		OTAUpgradeEntity otaPoint = pt.getOtaPoint();
		if(otaPoint == null){
			otaPoint = new OTAUpgradeEntity();
			pt.setOtaPoint(otaPoint);
		}
		try {
			if(StringUtil.isEmpty(pt.getAddrNo()))throw new RuntimeException("在协议层，OTA升级功能中,集中器编号为空！");
			if(DEBUG)log.info("OTA升级中....");
			/**当前版本与升级版本相同  不用升级*/
			if(terminal.getOtaVersion() != null && terminal.getOtaVersion().intValue() == pt.getHsId()){
				if(DEBUG) log.info("当前版本不需要更新");
				return;
			}
			/** 更新设备说明正在升级中*/
			terminal.setOtaStatus(OTAStatus.OTA_SEND.getId());
			terminal.setRemark(OTAStatus.OTA_SEND.getName());
			terminalDao.updateTemplateById(terminal);
			//下面说明是第移除 需要更新 升级进度
			if(otaPoint.getUpgradeVersion() == 0){
				otaPoint.setStartTime(System.currentTimeMillis());
				otaPoint.setUpgradeVersion(pt.getHsId());
				otaPoint.setOTAStaus(OTAStatus.OTA_RUNING);
				protocolRedisCache.updateOTAOPoint(pt.getAddrNo(),otaPoint);
			}
			if (pt.isUpdate()) {// 升级
				if(DEBUG)log.info("一条路走到黑....ota全流程走一遍....");
				sendUpgradeData(terminal,pt);
			} else {// 这里说明是传输数据 不升级
				if(terminal.getOtaVersion() == pt.getHsId()) if(DEBUG) log.info("当前升级包为最新的，不需要传输！");
				else{
					if(DEBUG) log.info("这里只发送数据，不升级");
					sendUpgradeData(terminal,pt);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			if(terminal != null){
				terminal.setOtaStatus(OTAStatus.OTA_FALL.getId());
				terminal.setRemark(e.getMessage());
				//端点 数据
				if(otaPoint != null){
					otaPoint.setOTAStaus(OTAStatus.OTA_FALL);
					otaPoint.setOtaFallMsg(e.getMessage());
				}
			}
			log.error("++++++++++OTA升级异常+++++++++,{}",e.getMessage());
			//e.printStackTrace();
		}finally{
			if(terminal != null){
				updateTerminal(terminal);
				protocolRedisCache.updateOTAOPoint(pt.getAddrNo(), pt.getOtaPoint());
			}
		}
	}
	@Async
	public void updateTerminal(CtkEquipTerminal terminal){
		terminalDao.updateTemplateById(terminal);
	}

	/**
	 * 发送数据后 是否升级
	 * */
	private void sendUpgradeData(CtkEquipTerminal terminal,ProtocolOTAEntity pt)
			throws Exception {
		if(pt.getBufferedIo() == null || pt.getBufferedIo().length < 1)RespResult.buidException("升级数据包有误");

		OTAUpgradeEntity otaPoint = pt.getOtaPoint();
		if(DEBUG)log.info("全新升级，不是端点升级");
		RespResult.resDataChange(askStorge(pt),"询问设备内存失败");

		List<byte[]> list = new ArrayList<byte[]>();
		RespResult.resDataChange(sendDataReady(list,pt),"通知设备开始传输数据失败");

		/*** 下面说明第一次发送数据  并且需要记住下数据长度  为了断点升级*/
		otaPoint.setTotalProgress(list.size());
		otaPoint.setOTAStaus(OTAStatus.OTA_SEND);
		protocolRedisCache.updateOTAOPoint(pt.getAddrNo(), otaPoint);//这个端点以后 可以给前端动态显示

		//下面发送数据包开始 Server传输给Client固件的指令 A2H
		RespResult.resDataChange(sendData(list,pt),"向硬件传输数据失败");

		//Server传输固件结束 A3H 发送数据结束
		RespResult.resDataChange(sendDataEnd(pt.getAddrNo()),"向设备，发送数据结束失败");

		/***Server询问Client固件校验是否正确 A4H*/
		RespResult.resDataChange(checkSendDataOk(pt),"询问硬件校验升级数据失败");
		/** 升级版本 */
		terminal.setOtaVersion(pt.getHsId());
		terminal.setRemark("发送数据包成功");
		//一路走到黑
		upgradeHandle(terminal,pt);
		/*if(pt.isUpdate()){//这里需升级
			//修改状态前
			//terminal.setOtaStatus(OTAStatus.OTA_RUNING.getId());
			//询问数据正确时，需将该状态该为  修改
			//terminalDao.updateTemplateById(terminal);
			upgradeHandle(terminal,pt);
		}else{
			//这里说明  发送数据成功
			terminal.setOtaStatus(OTAStatus.OTA_SUCCESS.getId());
		}*/
	}
	/**校验传输数据*/
	private RespWrapper<Boolean> checkSendDataOk(ProtocolOTAEntity pt)throws Exception {
		if(DEBUG)System.out.println("校验数据....");
		String sign = pt.getAddrNo()+ OperationCode.OTA_ASK_OK.getCode();
		OrderEntity entity = new OrderEntity(pt.getAddrNo(),
				OrderCodeType.CON_SUPER_LONG.getReqCode(),
				OperationCode.OTA_ASK_OK.getCode(), DEFAULT_BYTE);
		RespResult.resDataChange(retryOper(pt.getAddrNo(),entity,sign));
		if(DEBUG)System.out.println("【内部】：询问校验成功");
		return RespWrapper.makeSuccessResp(true);
	}

	/**发送数据结束指令*/
	private RespWrapper<Boolean> sendDataEnd(String addrNo)throws Exception {
		if(DEBUG)System.out.println("发送数据结束....");
		String sign = addrNo+OperationCode.OTA_SEND_END.getCode();
		OrderEntity entity = new OrderEntity(addrNo,
				OrderCodeType.CON_SUPER_LONG.getReqCode(),
				OperationCode.OTA_SEND_END.getCode(), DEFAULT_BYTE);
		RespResult.resDataChange(retryOper(addrNo,entity,sign));
		return RespWrapper.makeSuccessResp(true);
	}

	/***
	 * @Title: sendData
	 * @Description: TODO(发送数据包....)
	 * @param list   bl 是否 是端点升级   true 端点升级   false 全新升级
	 * @return
	 * boolean
	 * @throws
	 * @变更记录  2018年8月19日 上午10:36:56 zhaoLuYang create
	 */
	private RespWrapper<Boolean> sendData(List<byte[]> list,ProtocolOTAEntity pt)throws Exception {
		if(DEBUG) System.out.println("发送数据中....");
		String sign = pt.getAddrNo() + OperationCode.OTA_SEND_RUN.getCode();
		int size = list.size();
		byte[] encodeEvenProtocolID = SocketUtil.adjustValueConvert(size, DataLength.PAIR);//ProtocolIDUtil.encodeEvenProtocolID(convertBit(""+size, 4));
		int i = 0;
		OTAUpgradeEntity otaPoint = pt.getOtaPoint();
		byte[] bytes = null;
		byte[] encodeEvenProtocolID2 = null;
		if(DEBUG)System.out.println("总文件大小：【"+size+"】");
		for(;i<size;){
			if(DEBUG) {
				System.out.println(pt.getAddrNo()+"==>:【" + i + "】");
				/*if(i % 10 == 0)System.out.println("【" + i + "】");
				else System.out.print("【" + i + "】");*/
			}
			bytes = list.get(i);
			encodeEvenProtocolID2 = SocketUtil.adjustValueConvert(i+1, DataLength.PAIR);//ProtocolIDUtil.encodeEvenProtocolID(convertBit(""+i, 4));
			RespResult.resDataChange(sendDataSub(encodeEvenProtocolID,encodeEvenProtocolID2,bytes,sign,pt));
			otaPoint.setNewProgress(++i);
			/***记录下位置*/
			protocolRedisCache.updateOTAOPoint(pt.getAddrNo(), otaPoint);
		}
		return RespWrapper.makeSuccessResp(true);
	}
	//发送数据
	/**Server传输给Client固件的指令 A2H
	 */
	private RespWrapper<Boolean> sendDataSub(byte[] total,byte[] curNum,byte[] data,String sign,ProtocolOTAEntity pt)throws Exception{
		int lv1 = total.length;
		int lv2 = curNum.length;
		int lv3 = data.length;
		byte[] by = new byte[lv1+lv2+lv3];
		System.arraycopy(total, 0, by, 0, lv1);
		System.arraycopy(curNum, 0, by, lv1, lv2);
		System.arraycopy(data, 0, by, lv1+lv2, lv3);
		OrderEntity entity = new OrderEntity(pt.getAddrNo(),
				OrderCodeType.CON_SUPER_LONG.getReqCode(),
				OperationCode.OTA_SEND_RUN.getCode(), by);
		RespResult.resDataChange(retryOper(pt.getAddrNo(),entity,sign));
		return RespWrapper.makeSuccessResp(true);
	}
	/***  发送数据包前准备
	 *  Server通知Client即将开始固件传输 A1H
	 */
	//TODO
	private RespWrapper<Boolean> sendDataReady(List<byte[]> list,ProtocolOTAEntity pt)throws Exception{
		if(DEBUG)System.out.println("发送数据前....");
		String sign = pt.getAddrNo() + OperationCode.OTA_NOTICE_GO.getCode();
		Byte[] bufferedIo = pt.getBufferedIo();
		int size = bufferedIo.length;
		byte[] encodeEvenProtocolID = SocketUtil.adjustValueConvert(size, DataLength.FOUR);//   ProtocolIDUtil.encodeEvenProtocolID(convertBit(size+"",8));
		int num = size/LENGTH;
		boolean f = false;
		if(size % LENGTH != 0)f = true;
		for (int i = 0; i < num; i++) {
			byte[] by = new byte[LENGTH];
			for (int k = LENGTH * i; k < LENGTH * (1 + i); k++) {
				byte b = bufferedIo[k];
				by[k - LENGTH * i] = b;
			}
			list.add(by);
		}
		if(f){//说明还有一些数据没有完成
			int tmp = size-LENGTH*num;
			byte[] bt = new byte[tmp];
			for(int i=LENGTH*num;i<size;i++){
				byte b = bufferedIo[i];
				//t.add(b);
				bt[i - LENGTH*num] = b;
			}
			list.add(bt);
		}
		int lv1 = encodeEvenProtocolID.length;
		/*** CRC32校验*/
		byte[] encodeEvenProtocolID2 = CRC32Utils.encryptByte(bufferedIo);
		int lv2 = encodeEvenProtocolID2.length;
		byte[] bk = new byte[lv2+lv1];
		System.arraycopy(encodeEvenProtocolID, 0, bk, 0, lv1);
		System.arraycopy(encodeEvenProtocolID2, 0, bk, lv1, lv2);
		OrderEntity entity = new OrderEntity(pt.getAddrNo(),
				OrderCodeType.CON_SUPER_LONG.getReqCode(),
				OperationCode.OTA_NOTICE_GO.getCode(), bk);
		RespResult.resDataChange(retryOper(pt.getAddrNo(),entity,sign));
		return RespWrapper.makeSuccessResp(true);
	}

	/**询问存储内存*/
	private RespWrapper<Boolean> askStorge(ProtocolOTAEntity pt) throws Exception {
		if(DEBUG)System.out.println("询问存储大小...");
		String sign = pt.getAddrNo() + OperationCode.OTA_READ_STORGE.getCode();
		OrderEntity entity = new OrderEntity(pt.getAddrNo(),
				OrderCodeType.CON_SUPER_LONG.getReqCode(),
				OperationCode.OTA_READ_STORGE.getCode(), DEFAULT_BYTE);
		ResProtocol<Integer> integerResProtocol = RespResult.resDataChange(retryOper(pt.getAddrNo(), entity, sign));
		if (integerResProtocol.getData() > pt.getBufferedIo().length) {
			return RespWrapper.makeSuccessResp(true);
		}else{
			return RespWrapper.makeFailResp("设备存储小于升级包大小");
		}
	}

	/*** 通知设备升级    到这一步 */
	private void upgradeHandle(CtkEquipTerminal terminal,ProtocolOTAEntity pt) throws Exception {
		if(DEBUG)log.info("通知升级....");
		String addrNo = pt.getAddrNo();
		String sign = addrNo + OperationCode.OTA_UPDATE_START.getCode();
		// 通知设备升级
		OrderEntity entity = new OrderEntity(addrNo,OrderCodeType.CON_SUPER_LONG.getReqCode(),OperationCode.OTA_UPDATE_START.getCode(), DEFAULT_BYTE);
		RespResult.resDataChange(retryOper(addrNo,entity,sign),"通知升级");
		terminal.setOtaStatus(OTAStatus.OTA_RUNING.getId());
		terminal.setRemark(OTAStatus.OTA_RUNING.getName());
		pt.getOtaPoint().setOTAStaus(OTAStatus.OTA_SUCCESS);
		/**如果通知成功  需将集中器的状态改成掉线*/
		terminal.setStatus(TerminalStatusEnum.IN_MAINTEN.getId());
		/**强制设备掉线*/
		LocalProtocolCache.removeHearbeatValue(addrNo);
		LocalProtocolCache.closedChannelCache(addrNo);
		if(DEBUG) log.info("通知升级成功....");
	}
	/**重试*/
	private RespWrapper<ResProtocol<Integer>> retryOper(String addrNo,
			OrderEntity entity, String sign)throws Exception {
		/** 清空以前效应的数据*/
		try {
			for (int i = 0; i < RETRY_NUM; i++) {
				Channel channel = LocalProtocolCache.findChannelByTerminalNo(addrNo);
				if(channel == null) return RespWrapper.makeFailResp("根据集中编号获取连接设备socket失败！");
				channel.writeAndFlush(entity);
				ResProtocol<Integer> res =  getRedisListValue(sign);
				//TODO
				//说明失败  需要重试 以后这里需要调整
				if (res == null) continue;
				//如果失败 继续重试
				if(res.getAckCode() == AckCode.SUCCESS)return RespWrapper.makeSuccessResp(res);
			}
		}catch (Exception e){
			throw new RuntimeException(e.getMessage());
		}
		return RespWrapper.makeFailResp("重试【"+RETRY_NUM+"】次失败");
	}
	/*** 阻塞获取值*/
	private <V> V getRedisListValue(String key) {
		try {
			if(StringUtil.isNotEmpty(key))return (V)new FutureAdapter(key).get(10,TimeUnit.SECONDS);
		}catch (Exception e){
			e.printStackTrace();
		}
		return null;
	}




}
