package com.comtika.safe.core.weChat;

import com.alibaba.fastjson.JSONObject;
import com.comtika.safe.common.constant.ProjectConstant;
import com.comtika.safe.common.response.RespWrapper;
import com.comtika.safe.common.utils.StringUtil;
import com.comtika.safe.common.utils.encrypt.MD5Coding;
import com.comtika.safe.common.utils.http.HttpResult;
import com.comtika.safe.common.utils.http.HttpUtil;
import lombok.Data;
import org.omg.CORBA.SystemException;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.*;
import java.net.HttpURLConnection;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.*;

public class WeChatPayUtil {
    private static final String SYMBOLS = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";

    /**
     * 获取随机字符串 Nonce Str
     *
     * @return String 随机字符串
     */
    public static String generateNonceStr() {
        char[] nonceChars = new char[32];
        int size = SYMBOLS.length();
        for (int index = 0; index < nonceChars.length; ++index) {
            nonceChars[index] = SYMBOLS.charAt(ProjectConstant.RANDOM.nextInt(size));
        }
        return new String(nonceChars);
    }

    /**
     * 判断是否在微信中打开且微信版本号5.0及以上
     *
     * @param request
     * @return
     */
    public static boolean isWeiXin5(HttpServletRequest request) {
        String agent = request.getHeader("user-agent");
        String mm = "MicroMessenger";
        int index = agent.indexOf(mm);
        if (index != -1) {
            try {
                index += mm.length() + 1;
                String wx = agent.substring(index, index + 1);
                int d = Integer.parseInt(wx);
                if (d >= 5) {
                    return true;
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        return false;
    }

    /**
     * 返回处理结果给微信, 如:Util.write2wx(response, "SUCCESS", "OK")
     *
     * @param response
     * @param return_code
     * @param return_msg
     */
    public static void write2wx(HttpServletResponse response, String return_code, String return_msg) {
        StringBuilder xmlBuf = new StringBuilder();
        xmlBuf.append("<xml>\r\n");
        xmlBuf.append("\t<return_code>" + "<![CDATA[" + return_code + "]]>" + "</return_code>\r\n");
        xmlBuf.append("\t<return_msg>" + "<![CDATA[" + return_msg + "]]>" + "</return_msg>\r\n");
        xmlBuf.append("</xml>");

        String xml = xmlBuf.toString();
        try {
            PrintWriter out = response.getWriter();
            out.write(xml);
            out.flush();
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * 将数据集合生成微信签名,过程会自动过滤空值
     *
     * @param params  参数组
     * @param mch_key 微信商户平台密钥
     * @return
     */
    public static String createSign(Map<String, Object> params, String mch_key) {
        if (params == null) {
            return "";
        }
        //去掉空值和sign参数,使用Iterator遍历可在遍历同时删除元素
        Iterator<Map.Entry<String, Object>> it = params.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry<String, Object> en = it.next();
            String key = en.getKey();
            Object value = en.getValue();
            if (null == value || StringUtil.isEmpty(String.valueOf(value)) || key.equalsIgnoreCase("sign")) {
                it.remove();
            }
        }

        //按key排序后以[key=value&key=value]形式拼接,最后拼接微信秘钥,生成sign
        List<String> paramsKeys = new ArrayList<String>(params.keySet());
        Collections.sort(paramsKeys);

        StringBuilder sb = new StringBuilder();
        for (String param_key : paramsKeys) {
            String param_value = String.valueOf(params.get(param_key));
            sb.append(param_key).append("=").append(param_value).append("&");
        }
        sb.append("key").append("=").append(mch_key);

        //生成SIGN签名
        try {
            String signValue = MD5Coding.encode2HexStr(sb.toString().getBytes("UTF-8"));
            return signValue;
        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
            return "";
        }
    }


    /**
     * 将MAP参数组转成提交到微信的XML字符串
     *
     * @param params
     * @return
     */
    public static String mapFormatXML(Map<String, Object> params) {
        //按KEY排序
        List<String> keys = new ArrayList<String>(params.keySet());
        Collections.sort(keys);

        StringBuilder xmlBuf = new StringBuilder();
        xmlBuf.append("<xml>\r\n");

        for (String key : keys) {
            if (key.equalsIgnoreCase("sign")) {
                continue;        //将sign放到最后一行
            }
            String value = String.valueOf(params.get(key));
            String cdata_Value = "<![CDATA[" + value + "]]>";   //加上CDATA可转义特殊字符
            xmlBuf.append("\t<" + key + ">" + cdata_Value + "</" + key + ">\r\n");
        }
        xmlBuf.append("\t<sign>" + params.get("sign") + "</sign>\r\n");
        xmlBuf.append("</xml>");

        return xmlBuf.toString();
    }

    /**
     * POST数据到微信,获取结果
     *
     * @param wxOrderQueryUrl 微信API地址
     * @param postXmlString   POST提交的XML字符串
     * @return
     */
    public static RespWrapper<Map<String, Object>> postWX(String wxOrderQueryUrl, String postXmlString) {

        //调用微信,获取结果
        HttpURLConnection conn = null;
        try {
            conn = HttpUtil.createConnection(wxOrderQueryUrl);
        } catch (Exception e) {
            e.printStackTrace();
            return RespWrapper.makeFailResp("调用微信服务器,HTTP连接创建异常!");
        }
        HttpResult http_resp = HttpUtil.post(conn, postXmlString);
        if (http_resp.isStatusOK()) {
            String xml_result = http_resp.getResult();
            Map<String, Object> resultMap = WxXmlParser.parseXML(xml_result);        //解析微信响应的XML结果

            String return_code = String.valueOf(resultMap.get("return_code"));            //微信响应的通信状态码
            if (return_code.equalsIgnoreCase("SUCCESS")) {
                //通讯成功后,若需要即时校验微信结果是否签名匹配,防止数据被篡改,可在这里做处理.

                String result_code = String.valueOf(resultMap.get("result_code"));    /**微信响应的业务结果码*/
                if (result_code.equalsIgnoreCase("SUCCESS")) {
                    return RespWrapper.makeSuccessResp(resultMap);                        /**返回业务成功信息*/
                } else {
                    String err_code_des = String.valueOf(resultMap.get("err_code_des"));
                    return RespWrapper.makeFailResp(err_code_des);                        /**返回业务失败信息*/
                }
            } else {
                String return_msg = String.valueOf(resultMap.get("return_msg"));
                return RespWrapper.makeFailResp(return_msg);                            //返回通信失败信息
            }
        } else {
            return RespWrapper.makeFailResp("调用微信服务器未响应,状态[" + http_resp.getStatus() + "]");
        }
    }

    /**
     * 获取微信用户个人公开信息
     *
     * @param req
     * @param sessionKey
     * @return
     * @throws SystemException
     */
   /* public static RespWrapper<JSONObject> getWeChatUserPublicInfo(WeChatLoginReq req, String sessionKey) throws SystemException {
        // 验证签名有效--概率性签名校验失败：第一次授权时，屏蔽该校验-存在数据篡改风险
        String signature = HmacUtil.SHA1(req.getRawData() + sessionKey);
        if (!req.getSignature().equals(signature))
            return RespWrapper.makeFailResp("系统繁忙,请重试");
        try {
            // todo
            byte[] bytes = AESUtil.AES128CBC_PKCS7Decrypt_Base64(req.getEncryptedData(), sessionKey, req.getIv());
            if (null != bytes && bytes.length > 0) {
                String result = new String(bytes, StringUtil.ENC_UTF8);
                JSONObject userInfo = JSONObject.parseObject(result);
                return RespWrapper.makeSuccessResp(userInfo);
            }
        } catch (Exception e) {
            e.printStackTrace();
            return RespWrapper.makeFailResp("异常：" + e.getMessage());
        }
        return RespWrapper.makeFailResp("未知原因，异常");
    }*/

    /**
     * 获取微信授权token
     *
     * @param url 接口地址
     * @return
     */
    public static Map<String, Object> getJscodeSession(String url) {
        Map<String, Object> map = new HashMap<String, Object>();
        String result = HttpUtil.doGet(url);
        JSONObject jsonResult = JSONObject.parseObject(result);
        if (jsonResult.containsKey("session_key") && jsonResult.containsKey("openid")) {
            map.put("success", true);
            map.put("sessionKey", jsonResult.getString("session_key"));
            map.put("openId", jsonResult.getString("openid"));
        } else {
            map.put("success", false);
            map.put("errMsg", jsonResult.getIntValue("errcode") + "," + jsonResult.getString("errmsg"));
        }
        return map;
    }

    /**
     * 微信网页授权 -- 获取通行证封装
     *
     * @param url
     * @return
     */
    public static WeChatAccessToken getWebJscodeSession(String url) {
        String result = HttpUtil.doGet(url);
        JSONObject jsonResult = JSONObject.parseObject(result);
        WeChatAccessToken entity = new WeChatAccessToken();
        if (jsonResult.containsKey("access_token")) {
            entity.setSuccess(true);
            entity.setAccessToken(jsonResult.getString("access_token"));
            entity.setExpiresIn(jsonResult.getString("expires_in"));
            entity.setOpenid(jsonResult.getString("openid"));
            entity.setRefreshToken(jsonResult.getString("refresh_token"));
            entity.setScope(jsonResult.getString("scope"));
        } else {
            entity.setSuccess(false);
            entity.setErrorMsg(jsonResult.getIntValue("errcode") + "," + jsonResult.getString("errmsg"));
        }
        return entity;
    }

    @Data
    public static class WeChatAccessToken {
        /**
         * access_token	网页授权接口调用凭证,注意：此access_token与基础支持的access_token不同
         * expires_in	access_token接口调用凭证超时时间，单位（秒）
         * refresh_token	用户刷新access_token
         * openid	用户唯一标识，请注意，在未关注公众号时，用户访问公众号的网页，也会产生一个用户和公众号唯一的OpenID
         * scope	用户授权的作用域，使用逗号（,）分隔
         */
        private String accessToken;
        private String expiresIn;
        private String refreshToken;
        private String openid;
        private String scope;
        private boolean isSuccess;
        private String errorMsg;
    }

    /**
     * 微信XML结果解析器
     */
    public static class WxXmlParser {
        private static class PropertyCollector extends DefaultHandler {
            private LinkedList<String> keyStack = new LinkedList<String>();

            private LinkedList<StringBuilder> valueStack = new LinkedList<StringBuilder>();

            private HashMap<String, Object> result = new HashMap<String, Object>();

            @Override
            public void startElement(String uri, String localName, String qName, Attributes attributes) {
                this.keyStack.add(qName);
                this.valueStack.add(new StringBuilder());
                return;
            }

            @Override
            public void endElement(String uri, String localName, String qName) {
                String key = this.keyStack.removeLast();
                StringBuilder value = this.valueStack.removeLast();
                result.put(key, value.toString());
                if (valueStack.size() > 0)
                    this.valueStack.getLast().append('<').append(key).append('>')
                            .append(value).append("</").append(key).append('>');
                return;
            }

            @Override
            public void characters(char[] ch, int start, int length) {
                this.valueStack.getLast().append(ch, start, length);
                return;
            }

            public Map<String, Object> returnResult() {
                this.result.remove("xml"); // 解析类的一个bug.这里随便修正下，后期再改
                return (this.result);
            }
        }

        public static Map<String, Object> parseXML(String xmlString) {
            if (xmlString == null || xmlString.equals(""))
                throw (new IllegalArgumentException("argument xmlString must not be null."));
            try {
                SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
                InputStream source = new ByteArrayInputStream(xmlString.getBytes("UTF-8"));
                PropertyCollector pc = new PropertyCollector();
                parser.parse(source, pc);
                return (pc.returnResult());
            } catch (Exception ex) {
                throw (new RuntimeException(ex.getMessage(), ex));
            }
        }

        public static Map<String, Object> parseXML(InputStream xmlStream) {
            if (xmlStream == null)
                throw (new IllegalArgumentException("argument xmlStream must not be null."));
            try {
                SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
                PropertyCollector pc = new PropertyCollector();
                parser.parse(xmlStream, pc);
                return (pc.returnResult());
            } catch (Exception ex) {
                throw (new RuntimeException(ex.getMessage(), ex));
            }
        }
    }

    public static class HmacUtil {
        public static String SHA1(String str) {
            try {
                // 指定sha1算法
                MessageDigest digest = MessageDigest.getInstance("SHA-1");
                digest.update(str.getBytes("UTF-8"));
                // 获取字节数组
                byte messageDigest[] = digest.digest();
                // Create Hex String
                StringBuffer hexString = new StringBuffer();
                // 字节数组转换为 十六进制 数
                for (int i = 0; i < messageDigest.length; i++) {
                    String shaHex = Integer.toHexString(messageDigest[i] & 0xFF);
                    if (shaHex.length() < 2) {
                        hexString.append(0);
                    }
                    hexString.append(shaHex);
                }
                return hexString.toString().toLowerCase();
            } catch (UnsupportedEncodingException ue) {
                ue.printStackTrace();
                return "";
            } catch (NoSuchAlgorithmException e) {
                e.printStackTrace();
                return "";
            }
        }
    }

}
